package ru.dbotthepony.mc.otm.block.entity.cable

import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import net.minecraft.core.BlockPos
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.UNIVERSE_TICKS
import ru.dbotthepony.mc.otm.capability.receiveEnergy
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.RelativeSide
import ru.dbotthepony.mc.otm.core.shuffle
import ru.dbotthepony.mc.otm.graph.GraphNodeList
import ru.dbotthepony.mc.otm.onceServer
import java.util.*
import kotlin.collections.ArrayList
import kotlin.collections.HashSet
import kotlin.math.ln

class EnergyCableGraph : GraphNodeList<EnergyCableBlockEntity.Node, EnergyCableGraph>() {
	private val livelyNodes = HashSet<EnergyCableBlockEntity.Node>()
	private val livelyNodesList = ArrayList<EnergyCableBlockEntity.Node>()

	fun addLivelyNode(node: EnergyCableBlockEntity.Node) {
		when (contains(node)) {
			ContainsStatus.ABOUT_TO_BE_REMOVED, ContainsStatus.ABOUT_TO_BE_ADDED -> { } // do nothing
			ContainsStatus.DOES_NOT_BELONG -> throw IllegalArgumentException("$node does not belong to $this")
			ContainsStatus.CONTAINS -> {
				if (livelyNodes.add(node)) {
					livelyNodesList.add(node)
				}
			}
		}
	}

	private class CacheEntry {
		val paths = ArrayList<SegmentPath>(1)
		var saturated = false
	}

	// TODO: LRU cache?
	private val pathCache = HashMap<Pair<EnergyCableBlockEntity.Node, EnergyCableBlockEntity.Node>, CacheEntry>()

	private class SearchNode(
		val node: EnergyCableBlockEntity.Node,
		target: EnergyCableBlockEntity.Node,
		var parent: SearchNode? = null,
		previouslySeen: Boolean
	) : Comparable<SearchNode> {
		val heuristics: Double

		init {
			var heuristics = node.position.distSqr(target.position) * 0.0001

			if (!previouslySeen) {
				heuristics -= ln(node.segment.availableThroughput.coerceAtMost(Decimal.LONG_MAX_VALUE).toDouble())
			}

			this.heuristics = heuristics
		}

		override fun compareTo(other: SearchNode): Int {
			return heuristics.compareTo(other.heuristics)
		}
	}

	class SegmentPath(private val a: BlockPos, private val b: BlockPos) {
		val segments = HashSet<Segment>()
		private var shortCircuit = false
		private var lastTickTransfers = 0
		private var lastTick = 0

		val availableThroughput: Decimal
			get() = segments.minOfOrNull { it.availableThroughput } ?: Decimal.ZERO

		operator fun contains(node: EnergyCableBlockEntity.Node): Boolean {
			return segments.any { node in it }
		}

		fun transfer(amount: Decimal, simulate: Boolean, instantSnapshot: MutableMap<Segment, Decimal>): Decimal {
			if (!amount.isPositive || shortCircuit) {
				return Decimal.ZERO
			}

			if (lastTickTransfers++ >= 40000) {
				if (lastTick == UNIVERSE_TICKS) {
					shortCircuit = true
					LOGGER.warn("Cable path from $a to $b appears to be a short circuit, disabling.")
				} else {
					lastTick = UNIVERSE_TICKS
					lastTickTransfers = 0
				}
			}

			var min = amount

			for (segment in segments) {
				min = minOf(segment.transfer(amount, instantSnapshot), min)
				if (!min.isPositive) return Decimal.ZERO
			}

			if (!simulate)
				segments.forEach { it.transfer(min) }

			return min
		}

		fun refund(amount: Decimal, simulate: Boolean, instantSnapshot: MutableMap<Segment, Decimal>) {
			segments.forEach { it.refund(amount, simulate, instantSnapshot) }
		}

		fun remove() {
			segments.toTypedArray().forEach { it.remove(this) }
		}
	}

	// TODO: Actual segment logic (merging segments)
	class Segment() {
		constructor(node: EnergyCableBlockEntity.Node) : this() {
			nodes.add(node)
		}

		private val nodes = HashSet<EnergyCableBlockEntity.Node>()
		private val paths = HashSet<SegmentPath>()

		operator fun contains(node: EnergyCableBlockEntity.Node): Boolean {
			return node in nodes
		}

		var throughput = Decimal.ZERO
			private set

		var transferredLastTick = Decimal.ZERO
			private set

		val availableThroughput: Decimal get() {
			if (!throughputKnown) {
				throughput = nodes.maxOf { it.energyThroughput }
				throughputKnown = true
			}

			return throughput - transferredLastTick
		}

		private var throughputKnown = false
		private var lastTick = 0
		private var lastPoweredStatus = 0

		private fun updateBlockstates(): Boolean {
			if (lastTick + 2 < UNIVERSE_TICKS) {
				transferredLastTick = Decimal.ZERO
				lastTick = UNIVERSE_TICKS
			}

			val newState = transferredLastTick.isPositive
			val numState = if (newState) 1 else -1

			if (numState != lastPoweredStatus) {
				lastPoweredStatus = numState
				nodes.forEach { if (it.segment === this) it.updatePoweredState(newState) }
			}

			return newState
		}

		private var hasBlockstateTimer = false

		private fun blockstateUpdateStep() {
			hasBlockstateTimer = updateBlockstates()

			if (hasBlockstateTimer) {
				onceServer(20) {
					blockstateUpdateStep()
				}
			}
		}

		fun transfer(amount: Decimal, instantSnapshot: MutableMap<Segment, Decimal>): Decimal {
			if (lastTick != UNIVERSE_TICKS) {
				transferredLastTick = Decimal.ZERO
				lastTick = UNIVERSE_TICKS
			}

			val currentTransferred = instantSnapshot[this] ?: transferredLastTick

			if (currentTransferred >= throughput || !amount.isPositive) {
				return Decimal.ZERO
			}

			val new = minOf(currentTransferred + amount, throughput)
			val diff = new - currentTransferred
			instantSnapshot[this] = new
			return diff
		}

		fun transfer(amount: Decimal): Decimal {
			if (lastTick != UNIVERSE_TICKS) {
				transferredLastTick = Decimal.ZERO
				lastTick = UNIVERSE_TICKS
			}

			if (transferredLastTick >= throughput || !amount.isPositive) {
				return Decimal.ZERO
			}

			val new = minOf(transferredLastTick + amount, throughput)
			val diff = new - transferredLastTick
			transferredLastTick = new

			if (!hasBlockstateTimer) {
				blockstateUpdateStep()
			}

			return diff
		}

		fun refund(amount: Decimal, simulate: Boolean, instantSnapshot: MutableMap<Segment, Decimal>) {
			if (simulate) {
				if (this in instantSnapshot) {
					instantSnapshot[this] = maxOf(instantSnapshot[this]!! - amount, Decimal.ZERO)
				}
			} else {
				transferredLastTick = maxOf(transferredLastTick, Decimal.ZERO)
			}
		}

		fun remove(node: EnergyCableBlockEntity.Node) {
			check(nodes.remove(node)) { "Tried to remove node $node from segment $this, but that node does not belong to this segment" }

			if (nodes.isEmpty())
				throughput = Decimal.ZERO
			else
				throughput = nodes.maxOf { it.energyThroughput }
		}

		fun add(node: EnergyCableBlockEntity.Node) {
			check(nodes.add(node)) { "Tried to add node $node to segment $this, but we already have that node added" }
			throughput = maxOf(throughput, node.energyThroughput)
		}

		fun add(path: SegmentPath) {
			check(paths.add(path)) { "Path $path should already contain $this" }
			check(path.segments.add(this)) { "Path set and Segment disagree whenever $this is absent from $path" }

			if (!throughputKnown) {
				throughput = nodes.maxOf { it.energyThroughput }
				throughputKnown = true
			}
		}

		fun remove(path: SegmentPath) {
			check(paths.remove(path)) { "Path $path shouldn't contain $this" }
			check(path.segments.remove(this)) { "Path set and Segment disagree whenever $this is present in $path" }
		}

		// breaks "instant snapshots" of segments atm
		// shouldn't cause major gameplay issues though
		fun split(): List<Segment> {
			if (nodes.isEmpty()) {
				throw IllegalStateException("Empty segment somehow?")
			} else if (nodes.size == 1) {
				return listOf(this)
			} else {
				lastPoweredStatus = 0
				val list = ArrayList(nodes)
				val itr = list.iterator()
				itr.next()
				val result = ArrayList<Segment>()
				result.add(this)

				for (v in itr) {
					v.segment = Segment()
					v.segment.paths.addAll(paths)
					paths.forEach { it.segments.add(v.segment) }
					result.add(v.segment)
				}

				return result
			}
		}
	}

	fun invalidatePathCache() {
		pathCache.clear()
		nodes.forEach { it.onInvalidate() }
	}

	// isn't exactly A*, but greedy algorithm, which searched for locally optimal solutions because they lead to globally optimal ones
	private fun findPath(a: EnergyCableBlockEntity.Node, b: EnergyCableBlockEntity.Node, existing: Collection<SegmentPath>, threshold: Decimal): SegmentPath? {
		val seenTop = existing.any { a in it }

		if (seenTop && a.energyThroughput <= threshold) {
			return null
		}

		val openNodes = PriorityQueue<SearchNode>()
		val seenNodes = HashSet<EnergyCableBlockEntity.Node>()

		openNodes.add(SearchNode(a, b, null, seenTop))

		while (openNodes.isNotEmpty()) {
			val first = openNodes.remove()
			openNodes.remove(first)

			if (first.node === b) {
				// solution found
				val solution = ArrayList<EnergyCableBlockEntity.Node>()

				var last = first.parent
				solution.add(first.node)

				while (last != null) {
					solution.add(last.node)
					last = last.parent
				}

				val touchedSegments = ArrayList<Segment>()
				solution.forEach { touchedSegments.addAll(it.segment.split()) }
				val path = SegmentPath(a.blockEntity.blockPos, b.blockEntity.blockPos)
				touchedSegments.forEach { it.add(path) }

				return path
			} else {
				for (neighbour in first.node.neighboursView.values) {
					if (!seenNodes.add(neighbour) || !neighbour.canTraverse) continue
					val seen = existing.any { neighbour in it }
					if (seen && neighbour.energyThroughput <= threshold) continue
					openNodes.add(SearchNode(neighbour, b, first, seen))
				}
			}
		}

		return null
	}

	private fun getPath(a: EnergyCableBlockEntity.Node, b: EnergyCableBlockEntity.Node, energyToTransfer: Decimal): SegmentPath? {
		if (!a.canTraverse || !b.canTraverse)
			return null

		val list = pathCache.computeIfAbsent(a to b) { CacheEntry() }

		if (!list.saturated) {
			var maxThroughput = list.paths.maxOfOrNull { it.availableThroughput } ?: Decimal.ZERO

			while (maxThroughput < energyToTransfer && !list.saturated) {
				val find = findPath(a, b, list.paths, maxThroughput)

				if (find == null) {
					list.saturated
				} else {
					list.paths.add(find)
					maxThroughput = maxOf(maxThroughput, find.availableThroughput)
				}
			}
		}

		// TODO: Immediate load selection is bad because it can easily lead to oscillations and non optimal throughput
		//  without user having any clue why their shit work suboptimally
		return list.paths.maxByOrNull { it.availableThroughput }
	}

	override fun onNodeRemoved(node: EnergyCableBlockEntity.Node) {
		if (livelyNodes.remove(node)) {
			check(livelyNodesList.remove(node))
		}

		invalidatePathCache()
	}

	override fun onNodeAdded(node: EnergyCableBlockEntity.Node) {
		check(livelyNodes.add(node))
		livelyNodesList.add(node)
		invalidatePathCache()
	}

	fun receiveEnergy(howMuch: Decimal, simulate: Boolean, fromNode: EnergyCableBlockEntity.Node, fromSide: RelativeSide): Decimal {
		livelyNodesList.shuffle(fromNode.blockEntity.level!!.random)

		val itr = livelyNodesList.iterator()
		var received = Decimal.ZERO
		var residue = howMuch.coerceAtMost(fromNode.energyThroughput)
		val snapshot = Reference2ObjectOpenHashMap<Segment, Decimal>()

		for (node in itr) {
			var hit = false

			for (side in node.sides.values) {
				if (side.isEnabled) {
					if (fromNode === node && side.side === fromSide) {
						hit = true
						continue
					}

					side.neighbour.get()?.let {
						if (it !is EnergyCableBlockEntity.CableSide) {
							val path = getPath(fromNode, node, residue)
							hit = true

							if (path != null) {
								val thisReceived = it.receiveEnergy(path.transfer(residue, simulate, snapshot), simulate)
								received += thisReceived
								residue -= thisReceived
								if (!residue.isPositive) return received
							}
						}
					}
				}
			}

			if (!hit) {
				itr.remove()
				check(livelyNodes.remove(node))
			}
		}

		return received
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
